from amaranth import Array, Module, Cat, Mux, ShapeLike, Signal, Assert, unsigned
from amaranth.build import Platform
from amaranth.lib import wiring, data
from amaranth.lib.wiring import In, Out
from amaranth.lib.memory import Memory, ReadPort, WritePort
from amaranth.sim import Simulator
from amaranth.utils import ceil_log2


class RGBLayout(data.StructLayout):
    def __init__(self, r_bits, g_bits, b_bits):
        super().__init__(
            {
                "red": unsigned(r_bits),
                "green": unsigned(g_bits),
                "blue": unsigned(b_bits),
            }
        )


Rgb888Layout = RGBLayout(8, 8, 8)

Rgb111Layout = RGBLayout(1, 1, 1)


class Hub75Stream(wiring.Signature):
    """A Hub75E Driver for a single string of panels."""

    def __init__(self):
        super().__init__(
            {
                "latch": Out(1),
                "oe": Out(1),
                "addr": Out(5),
                "rgb0": Out(3),
                "rgb1": Out(3),
                "display_clk": Out(1),
            }
        )


class Hub75Ctrl(wiring.Signature):
    """Hub75E control/addressing signals.
    This is separated because certain designs keep these common
    between all panel outputs (e.g Colorlight 5A-75b boards).

    """

    def __init__(self):
        super().__init__(
            {
                "latch": Out(1),
                "oe": Out(1),
                "addr": Out(5),
                "display_clk": Out(1),
            }
        )


class Hub75Data(wiring.Signature):
    """Data lines for HUB75 displays. When combined with Hub75Ctrl, this forms a complete HUB75 interface.
    These are kept separate as some devices have a single set of control signals.
    """

    def __init__(self):
        super().__init__(
            {
                "rgb0": Out(Rgb111Layout),
                "rgb1": Out(Rgb111Layout),
            }
        )

class SwapBuffer(wiring.Component):
    """A pair of BRAMs for holdling line data that are swapped between using an external signal.

    Contains one write port, and one read port. They are attached to separate BRAMs to allow for
    one to be read at the same time another is being updated.

    Parameters:
        depth: i32, depth of each memory buffer
        shape: ShapeLike, the underlying shape of the memory.

    Signals:
        write_port: WritePort.Signature(width, shape)
        read_port: ReadPort.Signature(width, shape)
        selector: In(1)

    """

    def __init__(self, shape: ShapeLike, depth: int):
        super().__init__(
            {
                "selector": In(1),
                "read_port": In(
                    ReadPort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
                "write_port": In(
                    WritePort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
            }
        )
        self.data_shape = shape
        self.depth = depth

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        m.submodules.bram0 = self.bram0 = Memory(
            shape=self.data_shape, depth=self.depth, init=[]
        )
        m.submodules.bram1 = self.bram1 = Memory(
            shape=self.data_shape, depth=self.depth, init=[]
        )

        read0 = self.bram0.read_port()
        write0 = self.bram0.write_port()
        read1 = self.bram1.read_port()
        write1 = self.bram1.write_port()

        # for name, member in self.write_port.signature.members.items():
        #     m.d.comb += self.write_port.members[name].eq(Mux(self.selector, write0[name], write1[name]))
        #

        m.d.comb += [
            write0.addr.eq(self.write_port.addr),
            write1.addr.eq(self.write_port.addr),
            write0.en.eq(~self.selector),
            write1.en.eq(self.selector),
            # self.write_port.data.eq(Mux(self.selector, write0.data, write1.data)),
            write0.data.eq(self.write_port.data),
            write1.data.eq(self.write_port.data),

            read0.addr.eq(self.read_port.addr),
            read1.addr.eq(self.read_port.addr),
            read0.en.eq(~self.selector),
            read1.en.eq(self.selector),
            self.read_port.data.eq(Mux(self.selector, read1.data, read0.data)),
        ]

        return m


class Hub75StringDriver(wiring.Component):
    """A data driver for Hub75 panels. This accesses the line memory and feeds out the data"""

    bcm_select: In(3)
    done: Out(1)
    start: In(1)
    active_bank: In(1)  # the bank to use
    bram_port: Out(WritePort.Signature(
        addr_width=9, shape=Rgb888Layout
    )) # the other line to be swapped to.
    display_out: Out(Hub75Data())  # data signal output.

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        # add two memories
        m.submodules.bram0 = bram0 = Memory(shape=Rgb888Layout, depth=512, init=[])
        m.submodules.bram1 = bram1 = Memory(shape=Rgb888Layout, depth=512, init=[])

        # We use two brams here so we can swap between each - that way we can update the data
        # while displaying the other line. Switching is controlled by the coordinator.

        readports = Mux([bram0.read_port(), bram1.read_port()])
        active_readport = readports[self.active_bank]
        m.d.comb += active_readport.eq(readports[self.active_bank])
        writeports = Array([bram0.write_port(), bram1.write_port()])
        m.d.comb += self.bram_port.eq(writeports[~self.active_bank])

        # We want to set up the bram ports
        port0 = bram0.read_port()
        port1 = bram1.read_port()
        # the enable for these ports is based on active bank.
        m.d.comb += port0.en.eq(self.active_bank)
        m.d.comb += port1.en.eq(~self.active_bank)

        counter = Signal(32)
        m.d.sync += counter.eq(counter + 1)

        ram_rgb_slice = Cat(
            active_readport.data["red"].bit_select(self.bcm_shift, 1),
            active_readport.data["blue"].bit_select(self.bcm_shift, 1),
            active_readport.data["green"].bit_select(self.bcm_shift, 1),
        )

        pixnum = Signal(8, reset=127)
        pixrow = Signal(1, reset=0)
        m.d.comb += self.buf_addr.eq(Cat(pixrow, pixnum))

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    self.done.eq(0),
                    counter.eq(0),
                    pixnum.eq(127),
                    pixrow.eq(0),
                ]
                with m.If(self.start == 1):
                    m.d.sync += active_readport.en.eq(1)
                    m.next = "prefetch"
            with m.State("prefetch"):
                with m.If(counter == 0):
                    m.d.sync += pixrow.eq(0)
                with m.Elif(counter == 1):
                    m.d.sync += pixrow.eq(1)
                with m.Elif(counter == 2):
                    m.d.sync += self.display_out.rgb0.eq(ram_rgb_slice)
                with m.Elif(counter == 3):
                    m.d.sync += [self.display_out.rgb1.eq(ram_rgb_slice), counter.eq(0)]
                    m.next = "writerow"

        return m


def test_stringdriver():
    dut = Hub75StringDriver()
    sim = Simulator(dut)
    sim.add_clock(1e-6)

    async def testbench(ctx):
        pass

    sim.add_testbench(testbench)
    with sim.write_vcd("output.vcd"):
        sim.run_until(1e-6 * 1000)


class Hub75Coordinator(wiring.Component):
    """A shared-control hub75 driver"""

    def __init__(self, n_strings=1):
        self.n_strings = n_strings
        super().__init__()

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        return m


class Hub75EDriver(wiring.Component):
    """An optimized driver for hub75 strings.
    This version is faster than most implementations by merging the exposure
    period and the data-write period to happen simultaneously. As a result,
    the display can be brighter due to higher duty cycle.
    """

    start: In(1)
    done: Out(1)
    out: Out(Hub75Stream())
    buf_addr: Out(9)
    buf_data: In(36)

    row_depth = 128
    bit_depth = 8
    bcm_len = 32

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)
        bcm_shift: Signal = Signal(4, init=7)
        m.d.sync += counter.eq(counter + 1)
        # construct helper signals.
        ram_r = self.buf_data[16:23]
        ram_b = self.buf_data[8:15]
        ram_g = self.buf_data[0:7]

        ram_rgb_slice = Cat(
            ram_r.bit_select(bcm_shift, 1),
            ram_g.bit_select(bcm_shift, 1),
            ram_b.bit_select(bcm_shift, 1),
        )
        pixnum = Signal(8, reset=127)
        pixrow = Signal(1, reset=0)
        m.d.comb += self.buf_addr.eq(Cat(pixrow, pixnum))

        should_clock = counter < (128 * 2 + 1)
        should_expose = (counter < (32 << (bcm_shift + 1) + 1)).bool() & (
            bcm_shift != 7
        ).bool()

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    bcm_shift.eq(7),
                    counter.eq(0),
                    self.done.eq(0),
                    pixnum.eq(127),
                    pixrow.eq(0),
                ]
                with m.If(self.start == 1):
                    m.next = "prefetch"
            with m.State("prefetch"):
                with m.If(counter == 0):
                    m.d.sync += pixrow.eq(0)
                with m.Elif(counter == 1):
                    m.d.sync += pixrow.eq(1)
                with m.Elif(counter == 2):
                    m.d.sync += self.out.rgb0.eq(ram_rgb_slice)
                with m.Elif(counter == 3):
                    m.d.sync += [self.out.rgb1.eq(ram_rgb_slice), counter.eq(0)]
                    m.next = "writerow"
            with m.State("writerow"):
                # expose if we haven't done it for long enough yet.
                m.d.sync += self.out.oe.eq(~(should_expose))
                with m.If(should_clock):
                    # clock is high on entry
                    m.d.sync += self.out.display_clk.eq(counter[1] == 0)

                    with m.If(counter[0:1] == 0):
                        # rising edge of the clock
                        m.d.sync += pixrow.eq(0)

                    with m.If(counter[0:1] == 1):
                        m.d.sync += [
                            pixrow.eq(1),
                            self.out.rgb0.eq(ram_rgb_slice),
                        ]
                    with m.If(counter[0:1] == 2):
                        m.d.sync += [
                            pixnum.eq(pixnum - 1),
                            pixrow.eq(0),
                            self.out.rgb1.eq(ram_rgb_slice),
                        ]

                with m.Elif(~(should_expose)):
                    # we are done both feeding in the new data and exposing the previous.
                    m.d.sync += [counter.eq(0), self.out.display_clk.eq(0)]
                    m.next = "latchout"
            with m.State("latchout"):
                m.d.sync += [
                    pixnum.eq(127),
                    self.out.latch.eq(1),
                ]
                with m.If(counter > 3):
                    m.d.sync += self.out.latch.eq(0)
                    m.d.sync += counter.eq(0)
                    with m.If(bcm_shift == 0):
                        m.next = "finish"
                    with m.Else():
                        m.d.sync += bcm_shift.eq(bcm_shift - 1)
                        m.next = "prefetch"

            with m.State("finish"):
                m.d.sync += Assert(bcm_shift == 0, "finish without bcm shift 0")

                with m.If(counter < (32)):
                    m.d.sync += self.out.oe.eq(0)
                with m.Else():
                    m.d.sync += [self.out.oe.eq(1), self.done.eq(1)]
                    m.next = "init"

        return m


if __name__ == "__main__":
    m = Hub75EDriver()
    from amaranth.cli import main

    main(m)
